"""
Usage catalog factory, publish the usage information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from pathlib import Path
from typing import TypeVar

from hub.catalog_factories.usage.comnet_catalog import ComnetCatalog
from hub.catalog_factories.usage.nrcan_catalog import NrcanCatalog
from hub.catalog_factories.usage.eilat_catalog import EilatCatalog
from hub.helpers.utils import validate_import_export_type

Catalog = TypeVar('Catalog')


class UsageCatalogFactory:
  """
  Usage catalog factory class
  """
  def __init__(self, handler, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/usage')
    self._catalog_type = '_' + handler.lower()
    validate_import_export_type(UsageCatalogFactory, handler)
    self._path = base_path

  @property
  def _comnet(self):
    """
    Retrieve Comnet catalog
    """
    return ComnetCatalog(self._path)

  @property
  def _nrcan(self):
    """
    Retrieve NRCAN catalog
    """
    # nrcan retrieves the data directly from github
    return NrcanCatalog(self._path)

  @property
  def _eilat(self):
    """
    Retrieve Eilat catalog
    """
    return EilatCatalog(self._path)

  @property
  def catalog(self) -> Catalog:
    """
    Enrich the city given to the class using the class given handler
    :return: Catalog
    """
    return getattr(self, self._catalog_type, lambda: None)



"""
Location module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Location:
  """
  Location
  """
  def __init__(self, country, city, region_code, climate_reference_city_latitude, climate_reference_city_longitude):
    self._country = country
    self._city = city
    self._region_code = region_code
    self._climate_reference_city_latitude = climate_reference_city_latitude
    self._climate_reference_city_longitude = climate_reference_city_longitude

  @property
  def city(self):
    """
    Get city name
    """
    return self._city

  @property
  def country(self):
    """
    Get country code
    """
    return self._country

  @property
  def region_code(self):
    """
    Get region
    """
    return self._region_code

  @property
  def climate_reference_city_latitude(self):
    """
    Get climate-reference-city latitude
    """
    return self._climate_reference_city_latitude

  @property
  def climate_reference_city_longitude(self):
    """
    Get climate-reference-city longitude
    """
    return self._climate_reference_city_longitude



"""
Dictionaries module for Montreal system to hub energy generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class MontrealSystemToHubEnergyGenerationSystem:
  """
  Montreal's system to hub energy generation system class
  """
  def __init__(self):
    self._dictionary = {
      'Unitary air conditioner with baseboard heater fuel fired boiler': cte.BOILER,
      'Unitary air conditioner with baseboard heater electrical boiler': cte.BOILER,
      '4 pipe fan coils with fuel fired boiler': cte.BOILER,
      '4 pipe fan coils with electrical resistance water boiler': cte.BOILER,
      'Single zone packaged rooftop unit with fuel-fired furnace and baseboards and fuel boiler for acs': cte.BOILER,
      'Single zone packaged rooftop unit with electrical resistance furnace and baseboards and fuel boiler for acs': cte.BOILER,
      'Single zone make-up air unit with baseboard heating with fuel fired boiler': cte.BOILER,
      'Single zone make-up air unit with electrical baseboard heating and DHW with resistance': cte.BASEBOARD,
      'Multi-zone built-up system with baseboard heater hydronic with fuel fired boiler': cte.BOILER,
      'Multi-zone built-up system with electrical baseboard heater and electrical hot water': cte.BASEBOARD,
      'Unitary air conditioner air cooled DX with external condenser': cte.CHILLER,
      '4 pipe fan coils with water cooled, water chiller': cte.CHILLER,
      'Single zone packaged rooftop unit with air cooled DX': cte.CHILLER,
      'Single zone make-up air unit with air cooled DX': cte.CHILLER,
      'Multi-zone built-up system with water cooled, water chiller': cte.CHILLER,
      'PV system': cte.PHOTOVOLTAIC,
      'Multi-zone built-up system with heat pump for cooling': cte.CHILLER,
      'Multi-zone built-up system with heat pump for heat': cte.HEAT_PUMP
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
Plane module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import TypeVar
import numpy as np

Point = TypeVar('Point')


class Plane:
  """
  Plane class
  """

  def __init__(self, origin, normal):
    self._origin = origin
    self._normal = normal
    self._equation = None
    self._opposite_normal = None

  @property
  def origin(self) -> Point:
    """
    Get plane origin point
    :return: Point
    """
    return self._origin

  @property
  def normal(self):
    """
    Get plane normal [x, y, z]
    :return: np.ndarray
    """
    return self._normal

  @property
  def equation(self) -> (float, float, float, float):
    """
    Get the plane equation components Ax + By + Cz + D = 0
    :return: (A, B, C, D)
    """
    if self._equation is None:
      a = self.normal[0]
      b = self.normal[1]
      c = self.normal[2]
      d = -1 * self.origin.coordinates[0] * self.normal[0]
      d += -1 * self.origin.coordinates[1] * self.normal[1]
      d += -1 * self.origin.coordinates[2] * self.normal[2]
      self._equation = a, b, c, d
    return self._equation

  def distance_to_point(self, point):
    """
    Distance between the given point and the plane
    :return: float
    """
    p = point
    e = self.equation
    denominator = np.abs((p[0] * e[0]) + (p[1] * e[1]) + (p[2] * e[2]) + e[3])
    numerator = np.sqrt((e[0]**2) + (e[1]**2) + (e[2]**2))
    return float(denominator / numerator)

  @property
  def opposite_normal(self):
    """
    get plane normal in the opposite direction [x, y, z]
    :return: np.ndarray
    """
    if self._opposite_normal is None:
      coordinates = []
      for coordinate in self.normal:
        coordinates.append(-coordinate)
      self._opposite_normal = np.array(coordinates)
    return self._opposite_normal



"""
Monthly values module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2020 Project Author Pilar Monsalvete pilar_monsalvete@yahoo.es
"""

import hub.helpers.constants as cte


class MonthlyValues:
  """
  Monthly values class
  """
  @staticmethod
  def get_mean_values(values):
    """
    Calculates the mean values for each month from a list with hourly values
    :return: [float] x 12
    :param values: [float] x 8760
    """
    out = []
    if values is not None:
      hour = 0
      for month in cte.DAYS_A_MONTH:
        total = 0
        for _ in range(0, cte.DAYS_A_MONTH[month]):
          for _ in range(0, 24):
            total += values[hour] / 24 / cte.DAYS_A_MONTH[month]
            hour += 1
        out.append(total)
    return out

  @staticmethod
  def get_total_month(values):
    """
    Calculates the total value for each month
    :return: [float] x 12
    :param values: [float] x 8760
    """
    out = []
    if values is not None:
      hour = 0
      for month in cte.DAYS_A_MONTH:
        total = 0
        for _ in range(0, cte.DAYS_A_MONTH[month]):
          for _ in range(0, 24):
            total += values[hour]
            hour += 1
        out.append(total)
    return out



"""
ThermalOpening module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import uuid
from typing import TypeVar, Union
from hub.helpers.configuration_helper import ConfigurationHelper as ch

Polygon = TypeVar('Polygon')


class ThermalOpening:
  """
  ThermalOpening class
  """
  def __init__(self):
    self._id = None
    self._area = None
    self._conductivity = None
    self._frame_ratio = None
    self._g_value = None
    self._thickness = None
    self._overall_u_value = None
    self._hi = ch().convective_heat_transfer_coefficient_interior
    self._he = ch().convective_heat_transfer_coefficient_exterior
    self._construction_name = None

  @property
  def id(self):
    """
    Get thermal zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def area(self) -> Union[None, float]:
    """
    Get thermal opening area in square meters
    :return: None or float
    """
    return self._area

  @area.setter
  def area(self, value):
    """
    Set thermal opening area in square meters
    :param value: float
    """
    if value is not None:
      self._area = float(value)

  @property
  def conductivity(self) -> Union[None, float]:
    """
    Get thermal opening conductivity in W/mK
    :return: None or float
    """
    return self._conductivity

  @conductivity.setter
  def conductivity(self, value):
    """
    Set thermal opening conductivity in W/mK
    :param value: float
    """
    # The code to calculate overall_u_value is duplicated here and in thickness_m.
    # This ensures a more robust code that returns the overall_u_value regardless the order the parameters are read.
    if value is not None:
      self._conductivity = float(value)
      if self._overall_u_value is None and self.thickness is not None:
        h_i = self.hi
        h_e = self.he
        r_value = 1 / h_i + 1 / h_e + float(self.thickness) / float(self._conductivity)
        self._overall_u_value = 1 / r_value

  @property
  def frame_ratio(self) -> Union[None, float]:
    """
    Get thermal opening frame ratio
    :return: None or float
    """
    return self._frame_ratio

  @frame_ratio.setter
  def frame_ratio(self, value):
    """
    Set thermal opening frame ratio
    :param value: float
    """
    if value is not None:
      self._frame_ratio = float(value)

  @property
  def g_value(self) -> Union[None, float]:
    """
    Get thermal opening transmittance at normal incidence
    :return: None or float
    """
    return self._g_value

  @g_value.setter
  def g_value(self, value):
    """
    Set thermal opening transmittance at normal incidence
    :param value: float
    """
    if value is not None:
      self._g_value = float(value)

  @property
  def thickness(self) -> Union[None, float]:
    """
    Get thermal opening thickness in meters
    :return: None or float
    """
    return self._thickness

  @thickness.setter
  def thickness(self, value):
    """
    Set thermal opening thickness in meters
    :param value: float
    """
    # The code to calculate overall_u_value is duplicated here and in conductivity.
    # This ensures a more robust code that returns the overall_u_value regardless the order the parameters are read.
    if value is not None:
      self._thickness = float(value)
      if self._overall_u_value is None and self.conductivity is not None:
        h_i = self.hi
        h_e = self.he
        r_value = 1 / h_i + 1 / h_e + float(self._thickness) / float(self.conductivity)
        self._overall_u_value = 1 / r_value

  @property
  def overall_u_value(self) -> Union[None, float]:
    """
    Get thermal opening overall U-value in W/m2K
    :return: None or float
    """
    return self._overall_u_value

  @overall_u_value.setter
  def overall_u_value(self, value):
    """
    Set thermal opening overall U-value in W/m2K
    :param value: float
    """
    if value is not None:
      self._overall_u_value = float(value)

  @property
  def hi(self) -> Union[None, float]:
    """
    Get internal convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._hi

  @hi.setter
  def hi(self, value):
    """
    Set internal convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._hi = float(value)

  @property
  def he(self) -> Union[None, float]:
    """
    Get external convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._he

  @he.setter
  def he(self, value):
    """
    Set external convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._he = float(value)

  @property
  def construction_name(self):
    """
    Get thermal opening construction name
    """
    return self._construction_name

  @construction_name.setter
  def construction_name(self, value):
    """
    Set thermal opening construction name
    """
    self._construction_name = value



"""
Operational costs included in the catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.catalog_factories.data_models.cost.fuel import Fuel


class OperationalCost:
  """
  Operational cost class
  """
  def __init__(self, fuels, maintenance_heating, maintenance_cooling, maintenance_pv, co2):
    self._fuels = fuels
    self._maintenance_heating = maintenance_heating
    self._maintenance_cooling = maintenance_cooling
    self._maintenance_pv = maintenance_pv
    self._co2 = co2

  @property
  def fuels(self) -> List[Fuel]:
    """
    Get fuels listed in capital costs
    :return: [Fuel]
    """
    return self._fuels

  @property
  def maintenance_heating(self):
    """
    Get cost of maintaining the heating system in currency/W
    :return: float
    """
    return self._maintenance_heating

  @property
  def maintenance_cooling(self):
    """
    Get cost of maintaining the cooling system in currency/W
    :return: float
    """
    return self._maintenance_cooling

  @property
  def maintenance_pv(self):
    """
    Get cost of maintaining the PV system in currency/m2
    :return: float
    """
    return self._maintenance_pv

  @property
  def co2(self):
    """
    Get cost of CO2 emissions in currency/kgCO2
    :return: float
    """
    return self._co2

  def to_dictionary(self):
    """Class content to dictionary"""
    _fuels = []
    for _fuel in self.fuels:
      _fuels.append(_fuel.to_dictionary())
    content = {'Maintenance': {'fuels': _fuels,
                               'cost of maintaining the heating system [currency/W]': self.maintenance_heating,
                               'cost of maintaining the cooling system [currency/W]': self.maintenance_cooling,
                               'cost of maintaining the PV system [currency/W]': self.maintenance_pv,
                               'cost of CO2 emissions [currency/kgCO2]': self.co2
                               }
               }

    return content



"""
EnergySystemsFactory retrieve the energy system module for the given region
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete pilar.monsalvete@concordi.
Code contributors: Peter Yefi peteryefi@gmail.com
"""
from pathlib import Path

from hub.helpers.utils import validate_import_export_type
from hub.imports.energy_systems.montreal_custom_energy_system_parameters import MontrealCustomEnergySystemParameters


class EnergySystemsFactory:
  """
  EnergySystemsFactory class
  """

  def __init__(self, handler, city, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/energy_systems')
    self._handler = '_' + handler.lower()
    validate_import_export_type(EnergySystemsFactory, handler)
    self._city = city
    self._base_path = base_path

  def _montreal_custom(self):
    """
    Enrich the city by using montreal custom energy systems catalog information
    """
    MontrealCustomEnergySystemParameters(self._city).enrich_buildings()
    self._city.level_of_detail.energy_systems = 1
    for building in self._city.buildings:
      building.level_of_detail.energy_systems = 1

  def enrich(self):
    """
    Enrich the city given to the class using the class given handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
BuildingsCluster module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, TypeVar

from hub.city_model_structure.city_objects_cluster import CityObjectsCluster

CityObject = TypeVar('CityObject')


class BuildingsCluster(CityObjectsCluster):
  """
  BuildingsCluster(CityObjectsCluster) class
  """
  def __init__(self, name, city_objects):
    self._cluster_type = 'buildings'
    super().__init__(name, self._cluster_type, city_objects)
    self._name = name
    self._city_objects = city_objects

  @property
  def type(self):
    """
    Get cluster type
    :return: str
    """
    return self._cluster_type

  @property
  def city_objects(self) -> List[CityObject]:
    """
    Get the list of city objects conforming the cluster
    :return: [CityObject]
    """
    return self._city_objects



"""
User repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import datetime
import logging

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.helpers.auth import Auth
from hub.persistence.repository import Repository
from hub.persistence.models import User as Model, Application as ApplicationModel, UserRoles


class User(Repository):
  """
  User class
  """
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(User, cls).__new__(cls)
    return cls._instance

  def insert(self, name: str, password: str, role: UserRoles, application_id: int):
    """
    Inserts a new user
    :param name: username
    :param password: user password
    :param role: user rol [Admin or Hub_Reader]
    :param application_id: user application id
    :return: Identity id
    """
    try:
      user = self.get_by_name_and_application(name, application_id)
      if user is not None:
        raise SQLAlchemyError(f'A user named {user.name} already exists for that application')
    except TypeError:
      pass
    try:
      user = Model(name=name, password=Auth.hash_password(password), role=role, application_id=application_id)
      self.session.add(user)
      self.session.flush()
      self.session.commit()
      self.session.refresh(user)
      return user.id
    except SQLAlchemyError as err:
      logging.error('An error occurred while creating user %s', err)
      raise SQLAlchemyError from err

  def update(self, user_id: int, name: str, password: str, role: UserRoles):
    """
    Updates a user
    :param user_id: the id of the user to be updated
    :param name: the name of the user
    :param password: the password of the user
    :param role: the role of the user
    :return: None
    """
    try:
      self.session.query(Model).filter(Model.id == user_id).update({
        'name': name,
        'password': Auth.hash_password(password),
        'role': role,
        'updated': datetime.datetime.utcnow()
      })
      self.session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while updating user: %s', err)
      raise SQLAlchemyError from err

  def delete(self, user_id: int):
    """
    Deletes a user with the id
    :param user_id: the user id
    :return: None
    """
    try:
      self.session.query(Model).filter(Model.id == user_id).delete()
      self.session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while fetching user: %s', err)
      raise SQLAlchemyError from err

  def get_by_name_and_application(self, name: str, application_id: int) -> Model:
    """
    Fetch user based on the email address
    :param name: Username
    :param application_id: User application name
    :return: User matching the search criteria or None
    """
    try:
      user = self.session.execute(
        select(Model).where(Model.name == name, Model.application_id == application_id)
      ).first()
      return user[0]
    except SQLAlchemyError as err:
      logging.error('Error while fetching user by name and application: %s', err)
      raise SQLAlchemyError from err
    except TypeError as err:
      logging.error('Error while fetching user, empty result %s', err)
      raise TypeError from err

  def get_by_name_application_id_and_password(self, name: str, password: str, application_id: int) -> Model:
    """
    Fetch user based on the name, password and application id
    :param name: Username
    :param password: User password
    :param application_id: Application id
    :return: User
    """
    try:
      user = self.session.execute(
        select(Model).where(Model.name == name, Model.application_id == application_id)
      ).first()
      if user:
        if Auth.check_password(password, user[0].password):
          return user[0]
    except SQLAlchemyError as err:
      logging.error('Error while fetching user by name: %s', err)
      raise SQLAlchemyError from err
    raise ValueError('Unauthorized')

  def get_by_name_application_uuid_and_password(self, name: str, password: str, application_uuid: str) -> Model:
    """
    Fetch user based on the email and password
    :param name: Username
    :param password: User password
    :param application_uuid: Application uuid
    :return: User
    """
    try:
      application = self.session.execute(
        select(ApplicationModel).where(ApplicationModel.application_uuid == application_uuid)
      ).first()
      return self.get_by_name_application_id_and_password(name, password, application[0].id)
    except SQLAlchemyError as err:
      logging.error('Error while fetching user by name: %s', err)
      raise SQLAlchemyError from err
    except ValueError as err:
      raise ValueError from err



"""
Energy System catalog archetype
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List

from hub.catalog_factories.data_models.energy_systems.system import System


class Archetype:
  """
  Archetype class
  """
  def __init__(self, lod, name, systems):

    self._lod = lod
    self._name = name
    self._systems = systems

  @property
  def lod(self):
    """
    Get level of detail of the catalog
    :return: string
    """
    return self._lod

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return self._name

  @property
  def systems(self) -> List[System]:
    """
    Get list of equipments that compose the total energy system
    :return: [Equipment]
    """
    return self._systems

  def to_dictionary(self):
    """Class content to dictionary"""
    _systems = []
    for _system in self.systems:
      _systems.append(_system.to_dictionary())
    content = {'Archetype': {'name': self.name,
                             'level of detail': self.lod,
                             'systems': _systems
                             }
               }
    return content



"""
TestUsageFactory test and validate the city model structure usage parameters
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from unittest import TestCase

from hub.imports.geometry_factory import GeometryFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.usage_factory import UsageFactory
from hub.helpers.dictionaries import Dictionaries


class TestUsageFactory(TestCase):
  """
  TestUsageFactory TestCase
  """
  def setUp(self) -> None:
    """
    Configure test environment
    :return:
    """
    self._city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  def _check_buildings(self, city):
    for building in city.buildings:
      self.assertIsNotNone(building.name, 'building name is none')
      self.assertIsNotNone(building.type, 'building type is none')
      self.assertIsNotNone(building.volume, 'building volume is none')
      self.assertIsNotNone(building.detailed_polyhedron, 'building detailed polyhedron is none')
      self.assertIsNotNone(building.simplified_polyhedron, 'building simplified polyhedron is none')
      self.assertIsNotNone(building.surfaces, 'building surfaces is none')
      self.assertIsNotNone(building.centroid, 'building centroid is none')
      self.assertIsNotNone(building.max_height, 'building max_height is none')
      self.assertEqual(len(building.external_temperature), 0, 'building external temperature is calculated')
      self.assertEqual(len(building.global_horizontal), 0, 'building global horizontal is calculated')
      self.assertEqual(len(building.diffuse), 0, 'building diffuse is calculated')
      self.assertEqual(len(building.beam), 0, 'building beam is calculated')
      self.assertIsNotNone(building.lower_corner, 'building lower corner is none')
      self.assertEqual(len(building.sensors), 0, 'building sensors are assigned')
      self.assertIsNotNone(building.internal_zones, 'no internal zones created')
      self.assertIsNotNone(building.grounds, 'building grounds is none')
      self.assertIsNotNone(building.walls, 'building walls is none')
      self.assertIsNotNone(building.roofs, 'building roofs is none')
      for internal_zone in building.internal_zones:
        if internal_zone.usages is not None:
          self.assertTrue(len(internal_zone.usages) > 0, 'usage zones are not defined')
      self.assertIsNone(building.basement_heated, 'building basement_heated is not none')
      self.assertIsNone(building.attic_heated, 'building attic_heated is not none')
      self.assertIsNone(building.terrains, 'building terrains is not none')
      self.assertIsNotNone(building.year_of_construction, 'building year_of_construction is none')
      self.assertIsNotNone(building.function, 'building function is none')
      self.assertEqual(len(building.heating_demand), 0, 'building heating is not none')
      self.assertEqual(len(building.cooling_demand), 0, 'building cooling is not none')
      self.assertIsNotNone(building.eave_height, 'building eave height is none')
      self.assertIsNotNone(building.roof_type, 'building roof type is none')
      self.assertIsNotNone(building.floor_area, 'building floor_area is none')

  def _check_usage(self, usage):
    self.assertIsNotNone(usage.name, 'usage is none')
    self.assertIsNotNone(usage.percentage, 'usage percentage is none')
    self.assertIsNotNone(usage.hours_day, 'hours per day is none')
    self.assertIsNotNone(usage.days_year, 'days per year is none')
    self.assertIsNotNone(usage.thermal_control, 'thermal control is none')
    self.assertIsNotNone(usage.thermal_control.mean_heating_set_point, 'control heating set point is none')
    self.assertIsNotNone(usage.thermal_control.heating_set_back, 'control heating set back is none')
    self.assertIsNotNone(usage.thermal_control.mean_cooling_set_point, 'control cooling set point is none')

  def test_import_comnet(self):
    """
    Enrich the city with the usage information from comnet and verify it
    """
    file = 'pluto_building.gml'
    city = self._get_citygml(file)
    for building in city.buildings:
      building.function = Dictionaries().pluto_function_to_hub_function[building.function]

    UsageFactory('comnet', city).enrich()
    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        self.assertIsNot(len(internal_zone.usages), 0, 'no building usage defined')
        for usage in internal_zone.usages:
          self._check_usage(usage)
          self.assertIsNotNone(usage.mechanical_air_change, 'mechanical air change is none')
          self.assertIsNotNone(usage.thermal_control.heating_set_point_schedules,
                               'control heating set point schedule is none')
          self.assertIsNotNone(usage.thermal_control.cooling_set_point_schedules,
                               'control cooling set point schedule is none')
          self.assertIsNotNone(usage.occupancy, 'occupancy is none')
          occupancy = usage.occupancy
          self.assertIsNotNone(occupancy.occupancy_density, 'occupancy density is none')
          self.assertIsNotNone(occupancy.latent_internal_gain, 'occupancy latent internal gain is none')
          self.assertIsNotNone(occupancy.sensible_convective_internal_gain,
                               'occupancy sensible convective internal gain is none')
          self.assertIsNotNone(occupancy.sensible_radiative_internal_gain,
                               'occupancy sensible radiant internal gain is none')
          self.assertIsNotNone(occupancy.occupancy_schedules, 'occupancy schedule is none')
          self.assertIsNotNone(usage.lighting, 'lighting is none')
          lighting = usage.lighting
          self.assertIsNotNone(lighting.density, 'lighting density is none')
          self.assertIsNotNone(lighting.latent_fraction, 'lighting latent fraction is none')
          self.assertIsNotNone(lighting.convective_fraction, 'lighting convective fraction is none')
          self.assertIsNotNone(lighting.radiative_fraction, 'lighting radiant fraction is none')
          self.assertIsNotNone(lighting.schedules, 'lighting schedule is none')
          self.assertIsNotNone(usage.appliances, 'appliances is none')
          appliances = usage.appliances
          self.assertIsNotNone(appliances.density, 'appliances density is none')
          self.assertIsNotNone(appliances.latent_fraction, 'appliances latent fraction is none')
          self.assertIsNotNone(appliances.convective_fraction, 'appliances convective fraction is none')
          self.assertIsNotNone(appliances.radiative_fraction, 'appliances radiant fraction is none')
          self.assertIsNotNone(appliances.schedules, 'appliances schedule is none')
          self.assertIsNotNone(usage.thermal_control.hvac_availability_schedules,
                               'control hvac availability is none')
          self.assertIsNotNone(usage.domestic_hot_water.density, 'domestic hot water density is none')
          self.assertIsNotNone(usage.domestic_hot_water.service_temperature,
                               'domestic hot water service temperature is none')
          self.assertIsNotNone(usage.domestic_hot_water.schedules, 'domestic hot water schedules is none')

  def test_import_nrcan(self):
    """
    Enrich the city with the usage information from nrcan and verify it
    """
    file = 'test.geojson'
    file_path = (self._example_path / file).resolve()
    city = GeometryFactory('geojson',
                           path=file_path,
                           height_field='citygml_me',
                           year_of_construction_field='ANNEE_CONS',
                           function_field='CODE_UTILI',
                           function_to_hub=Dictionaries().montreal_function_to_hub_function).city

    ConstructionFactory('nrcan', city).enrich()
    UsageFactory('nrcan', city).enrich()
    self._check_buildings(city)
    for building in city.buildings:
      for internal_zone in building.internal_zones:
        if internal_zone.usages is not None:
          self.assertIsNot(len(internal_zone.usages), 0, 'no building usage defined')
          for usage in internal_zone.usages:
            self._check_usage(usage)
            self.assertIsNotNone(usage.mechanical_air_change, 'mechanical air change is none')
            self.assertIsNotNone(usage.thermal_control.heating_set_point_schedules,
                                 'control heating set point schedule is none')
            self.assertIsNotNone(usage.thermal_control.cooling_set_point_schedules,
                                 'control cooling set point schedule is none')
            self.assertIsNotNone(usage.occupancy, 'occupancy is none')
            occupancy = usage.occupancy
            self.assertIsNotNone(occupancy.occupancy_density, 'occupancy density is none')
            self.assertIsNotNone(occupancy.latent_internal_gain, 'occupancy latent internal gain is none')
            self.assertIsNotNone(occupancy.sensible_convective_internal_gain,
                                 'occupancy sensible convective internal gain is none')
            self.assertIsNotNone(occupancy.sensible_radiative_internal_gain,
                                 'occupancy sensible radiant internal gain is none')
            self.assertIsNotNone(occupancy.occupancy_schedules, 'occupancy schedule is none')
            self.assertIsNotNone(usage.lighting, 'lighting is none')
            lighting = usage.lighting
            self.assertIsNotNone(lighting.density, 'lighting density is none')
            self.assertIsNotNone(lighting.latent_fraction, 'lighting latent fraction is none')
            self.assertIsNotNone(lighting.convective_fraction, 'lighting convective fraction is none')
            self.assertIsNotNone(lighting.radiative_fraction, 'lighting radiant fraction is none')
            self.assertIsNotNone(lighting.schedules, 'lighting schedule is none')
            self.assertIsNotNone(usage.appliances, 'appliances is none')
            appliances = usage.appliances
            self.assertIsNotNone(appliances.density, 'appliances density is none')
            self.assertIsNotNone(appliances.latent_fraction, 'appliances latent fraction is none')
            self.assertIsNotNone(appliances.convective_fraction, 'appliances convective fraction is none')
            self.assertIsNotNone(appliances.radiative_fraction, 'appliances radiant fraction is none')
            self.assertIsNotNone(appliances.schedules, 'appliances schedule is none')
            self.assertIsNotNone(usage.thermal_control.hvac_availability_schedules,
                                 'control hvac availability is none')
            self.assertIsNotNone(usage.domestic_hot_water.peak_flow, 'domestic hot water peak flow is none')
            self.assertIsNotNone(usage.domestic_hot_water.service_temperature,
                                 'domestic hot water service temperature is none')
            self.assertIsNotNone(usage.domestic_hot_water.schedules, 'domestic hot water schedules is none')



"""
Usage catalog schedule
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from typing import Union, List


class Schedule:
  """
  Schedule class
  """

  def __init__(self, schedule_type, values, data_type, time_step, time_range, day_types):
    self._type = schedule_type
    self._values = values
    self._data_type = data_type
    self._time_step = time_step
    self._time_range = time_range
    self._day_types = day_types

  @property
  def type(self) -> Union[None, str]:
    """
    Get schedule type
    :return: None or str
    """
    return self._type

  @property
  def values(self):
    """
    Get schedule values
    :return: [Any]
    """
    return self._values

  @property
  def data_type(self) -> Union[None, str]:
    """
    Get schedule data type from:
    ['any_number', 'fraction', 'on_off', 'temperature', 'humidity', 'control_type']
    :return: None or str
    """
    return self._data_type

  @property
  def time_step(self) -> Union[None, str]:
    """
    Get schedule time step from:
    ['second', 'minute', 'hour', 'day', 'week', 'month']
    :return: None or str
    """
    return self._time_step

  @property
  def time_range(self) -> Union[None, str]:
    """
    Get schedule time range from:
    ['minute', 'hour', 'day', 'week', 'month', 'year']
    :return: None or str
    """
    return self._time_range

  @property
  def day_types(self) -> Union[None, List[str]]:
    """
    Get schedule day types, as many as needed from:
    ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday', 'holiday', 'winter_design_day',
    'summer_design_day']
    :return: None or [str]
    """
    return self._day_types

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Schedule': {'type': self.type,
                            'time range': self.time_range,
                            'time step': self.time_step,
                            'data type': self.data_type,
                            'day types': self.day_types,
                            'values': self.values}
               }
    return content



"""
Cooling and Heating peak loads module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import math

import hub.helpers.constants as cte
from hub.helpers.peak_calculation.loads_calculation import LoadsCalculation

_MONTH_STARTING_HOUR = [0, 744, 1416, 2160, 2880, 3624, 4344, 5088, 5832, 6552, 7296, 8016, math.inf]


class PeakLoads:
  """
  PeakLoads class
  """
  def __init__(self, building=None):
    self._building = building

  def _can_be_calculated(self):
    levels_of_detail = self._building.level_of_detail
    can_be_calculated = True
    if levels_of_detail.geometry is None:
      can_be_calculated = False
    if levels_of_detail.geometry < 1:
      can_be_calculated = False
    if levels_of_detail.construction is None:
      can_be_calculated = False
    if levels_of_detail.construction < 1:
      can_be_calculated = False
    if levels_of_detail.usage is None:
      can_be_calculated = False
    if levels_of_detail.usage < 1:
      can_be_calculated = False
    if levels_of_detail.weather is None:
      can_be_calculated = False
    if levels_of_detail.weather < 2:
      can_be_calculated = False
    if levels_of_detail.surface_radiation is None:
      can_be_calculated = False
    if levels_of_detail.surface_radiation < 2:
      can_be_calculated = False
    return can_be_calculated

  @staticmethod
  def peak_loads_from_hourly(hourly_values):
    """
    Get peak loads from hourly
    :return: [int]
    """
    month = 1
    peaks = [0 for _ in range(12)]
    for i in range(0, len(hourly_values)):
      if _MONTH_STARTING_HOUR[month] <= i:
        month += 1
      if hourly_values[i] > peaks[month-1]:
        peaks[month-1] = hourly_values[i]
    return peaks

  @property
  def heating_peak_loads_from_methodology(self):
    """
    Get heating peak loads by calculate
    :return: [int]
    """
    if not self._can_be_calculated():
      return None
    monthly_heating_loads = []
    ambient_temperature = self._building.external_temperature[cte.HOUR]
    for month in range(0, 12):
      ground_temperature = self._building.ground_temperature[cte.MONTH]['2'][month]
      start_hour = _MONTH_STARTING_HOUR[month]
      end_hour = 8760
      if month < 11:
        end_hour = _MONTH_STARTING_HOUR[month + 1]
      heating_ambient_temperature = min(ambient_temperature[start_hour:end_hour])
      loads = LoadsCalculation(self._building)
      heating_load_transmitted = loads.get_heating_transmitted_load(heating_ambient_temperature, ground_temperature)
      heating_load_ventilation_sensible = loads.get_heating_ventilation_load_sensible(heating_ambient_temperature)
      # todo: include heating ventilation latent
      heating_load_ventilation_latent = 0
      heating_load = heating_load_transmitted + heating_load_ventilation_sensible + heating_load_ventilation_latent
      heating_load = max(heating_load, 0)
      monthly_heating_loads.append(heating_load)
    return monthly_heating_loads

  @property
  def cooling_peak_loads_from_methodology(self):
    """
    Get cooling peak loads by calculate
    :return: [int]
    """
    if not self._can_be_calculated():
      return None
    monthly_cooling_loads = []
    ambient_temperature = self._building.external_temperature[cte.HOUR]
    for month in range(0, 12):
      ground_temperature = self._building.ground_temperature[cte.MONTH]['2'][month]
      cooling_ambient_temperature = -100
      cooling_calculation_hour = -1
      start_hour = _MONTH_STARTING_HOUR[month]
      end_hour = 8760
      if month < 11:
        end_hour = _MONTH_STARTING_HOUR[month + 1]
      for hour in range(start_hour, end_hour):
        temperature = ambient_temperature[hour]
        if temperature > cooling_ambient_temperature:
          cooling_ambient_temperature = temperature
          cooling_calculation_hour = hour
      loads = LoadsCalculation(self._building)
      cooling_load_transmitted = loads.get_cooling_transmitted_load(cooling_ambient_temperature, ground_temperature)
      cooling_load_renovation_sensible = loads.get_cooling_ventilation_load_sensible(cooling_ambient_temperature)
      cooling_load_internal_gains_sensible = loads.get_internal_load_sensible()
      cooling_load_radiation = loads.get_radiation_load(cooling_calculation_hour)
      cooling_load_sensible = cooling_load_transmitted + cooling_load_renovation_sensible - cooling_load_radiation \
                              - cooling_load_internal_gains_sensible

      cooling_load_latent = 0
      cooling_load = cooling_load_sensible + cooling_load_latent
      cooling_load = min(cooling_load, 0)
      monthly_cooling_loads.append(abs(cooling_load))
    return monthly_cooling_loads



"""
Cost item properties
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union


class ItemDescription:
  """
  Item description class
  """
  def __init__(self, item_type,
               initial_investment=None,
               initial_investment_unit=None,
               refurbishment=None,
               refurbishment_unit=None,
               reposition=None,
               reposition_unit=None,
               lifetime=None):

    self._item_type = item_type
    self._initial_investment = initial_investment
    self._initial_investment_unit = initial_investment_unit
    self._refurbishment = refurbishment
    self._refurbishment_unit = refurbishment_unit
    self._reposition = reposition
    self._reposition_unit = reposition_unit
    self._lifetime = lifetime

  @property
  def type(self):
    """
    Get item type
    :return: str
    """
    return self._item_type

  @property
  def initial_investment(self) -> Union[tuple[None, None], tuple[float, str]]:
    """
    Get initial investment of the specific item in given units
    :return: None, None or float, str
    """
    return self._initial_investment, self._initial_investment_unit

  @property
  def refurbishment(self) -> Union[tuple[None, None], tuple[float, str]]:
    """
    Get refurbishment costs of the specific item in given units
    :return: None, None or float, str
    """
    return self._refurbishment, self._refurbishment_unit

  @property
  def reposition(self) -> Union[tuple[None, None], tuple[float, str]]:
    """
    Get reposition costs of the specific item in given units
    :return: None, None or float, str
    """
    return self._reposition, self._reposition_unit

  @property
  def lifetime(self) -> Union[None, float]:
    """
    Get lifetime in years
    :return: None or float
    """
    return self._lifetime

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Item': {'type': self.type,
                        'initial investment': self.initial_investment[0],
                        'initial investment units': self.initial_investment[1],
                        'refurbishment': self.refurbishment[0],
                        'refurbishment units': self.refurbishment[1],
                        'reposition': self.reposition[0],
                        'reposition units': self.reposition[1],
                        'life time [years]': self.lifetime
                        }
               }

    return content



"""
Point module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import math


class Point:
  """
  Point class
  """

  def __init__(self, coordinates):
    self._coordinates = coordinates

  @property
  def coordinates(self):
    """
    Get point coordinates
    :return: [ndarray]
    """
    return self._coordinates

  def distance_to_point(self, other_point):
    """
    Calculates distance between points in an n-D Euclidean space
    :param other_point: point or vertex
    :return: float
    """
    power = 0
    for dimension in enumerate(self.coordinates):
      power += math.pow(other_point.coordinates[dimension]-self.coordinates[dimension], 2)
    distance = math.sqrt(power)
    return distance



"""Definition of meta model 'greenerycatalog'."""
from functools import partial
import pyecore.ecore as Ecore
from pyecore.ecore import *


name = 'greenerycatalog'
nsURI = 'http://ca.concordia/greenerycatalog'
nsPrefix = 'greenery'

eClass = EPackage(name=name, nsURI=nsURI, nsPrefix=nsPrefix)

eClassifiers = {}
getEClassifier = partial(Ecore.getEClassifier, searchspace=eClassifiers)
Management = EEnum('Management', literals=['Intensive', 'Extensive', 'SemiIntensive', 'NA'])

Roughness = EEnum('Roughness', literals=['VeryRough', 'Rough',
                                         'MediumRough', 'MediumSmooth', 'Smooth', 'VerySmooth'])


class Soil(EObject, metaclass=MetaEClass):

  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  roughness = EAttribute(eType=Roughness, unique=True, derived=False,
                         changeable=True, default_value=Roughness.MediumRough)
  conductivityOfDrySoil = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='1.0 W/(m*K)')
  densityOfDrySoil = EAttribute(eType=EString, unique=True, derived=False,
                                changeable=True, default_value='1100 kg/m³')
  specificHeatOfDrySoil = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='1200 J/(kg*K)')
  thermalAbsorptance = EAttribute(eType=EString, unique=True,
                                  derived=False, changeable=True, default_value='0.9')
  solarAbsorptance = EAttribute(eType=EString, unique=True,
                                derived=False, changeable=True, default_value='0.7')
  visibleAbsorptance = EAttribute(eType=EString, unique=True,
                                  derived=False, changeable=True, default_value='0.75')
  saturationVolumetricMoistureContent = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='0.0')
  residualVolumetricMoistureContent = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='0.05')
  initialVolumetricMoistureContent = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='0.1')

  def __init__(self, *, name=None, roughness=None, conductivityOfDrySoil=None, densityOfDrySoil=None, specificHeatOfDrySoil=None, thermalAbsorptance=None, solarAbsorptance=None, visibleAbsorptance=None, saturationVolumetricMoistureContent=None, residualVolumetricMoistureContent=None, initialVolumetricMoistureContent=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if roughness is not None:
      self.roughness = roughness

    if conductivityOfDrySoil is not None:
      self.conductivityOfDrySoil = conductivityOfDrySoil

    if densityOfDrySoil is not None:
      self.densityOfDrySoil = densityOfDrySoil

    if specificHeatOfDrySoil is not None:
      self.specificHeatOfDrySoil = specificHeatOfDrySoil

    if thermalAbsorptance is not None:
      self.thermalAbsorptance = thermalAbsorptance

    if solarAbsorptance is not None:
      self.solarAbsorptance = solarAbsorptance

    if visibleAbsorptance is not None:
      self.visibleAbsorptance = visibleAbsorptance

    if saturationVolumetricMoistureContent is not None:
      self.saturationVolumetricMoistureContent = saturationVolumetricMoistureContent

    if residualVolumetricMoistureContent is not None:
      self.residualVolumetricMoistureContent = residualVolumetricMoistureContent

    if initialVolumetricMoistureContent is not None:
      self.initialVolumetricMoistureContent = initialVolumetricMoistureContent


class Plant(EObject, metaclass=MetaEClass):

  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  height = EAttribute(eType=EString, unique=True, derived=False,
                      changeable=True, default_value='0.1 m')
  leafAreaIndex = EAttribute(eType=EString, unique=True, derived=False,
                             changeable=True, default_value='2.5')
  leafReflectivity = EAttribute(eType=EString, unique=True,
                                derived=False, changeable=True, default_value='0.1')
  leafEmissivity = EAttribute(eType=EString, unique=True, derived=False,
                              changeable=True, default_value='0.9')
  minimalStomatalResistance = EAttribute(
    eType=EString, unique=True, derived=False, changeable=True, default_value='100.0  s/m')
  co2Sequestration = EAttribute(eType=EString, unique=True, derived=False,
                                changeable=True, default_value='kgCO₂eq')
  growsOn = EReference(ordered=True, unique=True, containment=False, derived=False, upper=-1)

  def __init__(self, *, name=None, height=None, leafAreaIndex=None, leafReflectivity=None, leafEmissivity=None, minimalStomatalResistance=None, growsOn=None, co2Sequestration=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if height is not None:
      self.height = height

    if leafAreaIndex is not None:
      self.leafAreaIndex = leafAreaIndex

    if leafReflectivity is not None:
      self.leafReflectivity = leafReflectivity

    if leafEmissivity is not None:
      self.leafEmissivity = leafEmissivity

    if minimalStomatalResistance is not None:
      self.minimalStomatalResistance = minimalStomatalResistance

    if co2Sequestration is not None:
      self.co2Sequestration = co2Sequestration

    if growsOn:
      self.growsOn.extend(growsOn)


class SupportEnvelope(EObject, metaclass=MetaEClass):

  roughness = EAttribute(eType=Roughness, unique=True, derived=False,
                         changeable=True, default_value=Roughness.MediumRough)
  solarAbsorptance = EAttribute(eType=EDouble, unique=True,
                                derived=False, changeable=True, default_value=0.0)
  conductivity = EAttribute(eType=EDouble, unique=True, derived=False,
                            changeable=True, default_value=0.0)
  visibleAbsorptance = EAttribute(eType=EDouble, unique=True,
                                  derived=False, changeable=True, default_value=0.0)
  specificHeat = EAttribute(eType=EDouble, unique=True, derived=False,
                            changeable=True, default_value=0.0)
  density = EAttribute(eType=EDouble, unique=True, derived=False,
                       changeable=True, default_value=0.0)
  thermalAbsorptance = EAttribute(eType=EDouble, unique=True,
                                  derived=False, changeable=True, default_value=0.0)

  def __init__(self, *, roughness=None, solarAbsorptance=None, conductivity=None, visibleAbsorptance=None, specificHeat=None, density=None, thermalAbsorptance=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if roughness is not None:
      self.roughness = roughness

    if solarAbsorptance is not None:
      self.solarAbsorptance = solarAbsorptance

    if conductivity is not None:
      self.conductivity = conductivity

    if visibleAbsorptance is not None:
      self.visibleAbsorptance = visibleAbsorptance

    if specificHeat is not None:
      self.specificHeat = specificHeat

    if density is not None:
      self.density = density

    if thermalAbsorptance is not None:
      self.thermalAbsorptance = thermalAbsorptance


class GreeneryCatalog(EObject, metaclass=MetaEClass):

  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  description = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  source = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  plantCategories = EReference(ordered=True, unique=True,
                               containment=True, derived=False, upper=-1)
  vegetationCategories = EReference(ordered=True, unique=True,
                                    containment=True, derived=False, upper=-1)
  soils = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

  def __init__(self, *, name=None, description=None, source=None, plantCategories=None, vegetationCategories=None, soils=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if description is not None:
      self.description = description

    if source is not None:
      self.source = source

    if plantCategories:
      self.plantCategories.extend(plantCategories)

    if vegetationCategories:
      self.vegetationCategories.extend(vegetationCategories)

    if soils:
      self.soils.extend(soils)


class PlantCategory(EObject, metaclass=MetaEClass):
  """Excluding (that is non-overlapping) categories like Trees, Hedeges, Grasses that help users finding a specific biol. plant species."""
  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  plants = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

  def __init__(self, *, name=None, plants=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if plants:
      self.plants.extend(plants)


class IrrigationSchedule(EObject, metaclass=MetaEClass):

  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)

  def __init__(self, *, name=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name


class Vegetation(EObject, metaclass=MetaEClass):
  """Plant life or total plant cover (as of an area)"""
  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  thicknessOfSoil = EAttribute(eType=EString, unique=True, derived=False,
                               changeable=True, default_value='20 cm')
  management = EAttribute(eType=Management, unique=True, derived=False,
                          changeable=True, default_value=Management.NA)
  airGap = EAttribute(eType=EString, unique=True, derived=False,
                      changeable=True, default_value='0.0 cm')
  soil = EReference(ordered=True, unique=True, containment=False, derived=False)
  plants = EReference(ordered=True, unique=True, containment=True, derived=False, upper=-1)

  def __init__(self, *, name=None, thicknessOfSoil=None, soil=None, plants=None, management=None, airGap=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if thicknessOfSoil is not None:
      self.thicknessOfSoil = thicknessOfSoil

    if management is not None:
      self.management = management

    if airGap is not None:
      self.airGap = airGap

    if soil is not None:
      self.soil = soil

    if plants:
      self.plants.extend(plants)


class VegetationCategory(EObject, metaclass=MetaEClass):
  """Excluding (that is non-overlapping) categories to help users finding a specific vegetation template."""
  name = EAttribute(eType=EString, unique=True, derived=False, changeable=True)
  vegetationTemplates = EReference(ordered=True, unique=True,
                                   containment=True, derived=False, upper=-1)

  def __init__(self, *, vegetationTemplates=None, name=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if name is not None:
      self.name = name

    if vegetationTemplates:
      self.vegetationTemplates.extend(vegetationTemplates)


class PlantPercentage(EObject, metaclass=MetaEClass):

  percentage = EAttribute(eType=EString, unique=True, derived=False,
                          changeable=True, default_value='100')
  plant = EReference(ordered=True, unique=True, containment=False, derived=False)

  def __init__(self, *, percentage=None, plant=None):
    # if kwargs:
    #    raise AttributeError('unexpected arguments: {}'.format(kwargs))

    super().__init__()

    if percentage is not None:
      self.percentage = percentage

    if plant is not None:
      self.plant = plant



"""
TestCityMerge test and validate the merge of several cities into one
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import copy
import distutils.spawn
import subprocess
from pathlib import Path
from unittest import TestCase

from hub.city_model_structure.city import City
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.results_factory import ResultFactory
from hub.exports.exports_factory import ExportsFactory
import hub.helpers.constants as cte


class TestCityMerge(TestCase):
  """
  Functional TestCityMerge
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()
    self._executable = 'sra'

  def test_merge(self):
    file_path = Path(self._example_path / 'test.geojson').resolve()
    full_city = GeometryFactory('geojson', file_path, height_field='citygml_me').city
    self.assertEqual(17, len(full_city.buildings), 'Wrong number of buildings')
    odd_city = City(full_city.lower_corner, full_city.upper_corner, full_city.srs_name)
    even_city = City(full_city.lower_corner, full_city.upper_corner, full_city.srs_name)
    for building in full_city.buildings:
      if int(building.name) % 2 == 0:
        even_city.add_city_object(copy.deepcopy(building))
      else:
        odd_city.add_city_object(copy.deepcopy(building))
    self.assertEqual(8, len(odd_city.buildings), 'Wrong number of odd buildings')
    self.assertEqual(9, len(even_city.buildings), 'Wrong number of par buildings')
    merged_city = odd_city.merge(even_city)
    self.assertEqual(17, len(merged_city.buildings), 'Wrong number of buildings in merged city')
    merged_city = even_city.merge(odd_city)
    self.assertEqual(17, len(merged_city.buildings), 'Wrong number of buildings in merged city')
    merged_city = full_city.merge(odd_city).merge(even_city)
    self.assertEqual(17, len(merged_city.buildings), 'Wrong number of buildings in merged city')

  def test_merge_with_radiation(self):
    sra = distutils.spawn.find_executable('sra')
    file_path = Path(self._example_path / 'test.geojson').resolve()

    full_city = GeometryFactory('geojson', file_path, height_field='citygml_me').city
    even_city = City(full_city.lower_corner, full_city.upper_corner, full_city.srs_name)
    for building in full_city.buildings:
      if int(building.name) % 2 == 0:
        even_city.add_city_object(copy.deepcopy(building))
    ExportsFactory('sra', full_city, self._output_path).export()
    sra_file = str((self._output_path / f'{full_city.name}_sra.xml').resolve())
    subprocess.run([sra, sra_file], stdout=subprocess.DEVNULL)
    ResultFactory('sra', full_city, self._output_path).enrich()
    self.assertEqual(17, len(full_city.buildings), 'Wrong number of buildings')
    merged_city = full_city.merge(even_city)

    full_city_building_total_radiation = 0
    for building in merged_city.buildings:
      for surface in building.surfaces:
        if surface.global_irradiance:
          full_city_building_total_radiation += surface.global_irradiance[cte.YEAR][0]

    merged_city_building_total_radiation = 0
    for building in merged_city.buildings:
      for surface in building.surfaces:
        if surface.global_irradiance:
          merged_city_building_total_radiation += surface.global_irradiance[cte.YEAR][0]
    self.assertEqual(full_city_building_total_radiation, merged_city_building_total_radiation)

    merged_city = even_city.merge(full_city)
    merged_city_building_total_radiation = 0
    for building in merged_city.buildings:
      for surface in building.surfaces:
        if surface.global_irradiance:
          merged_city_building_total_radiation += surface.global_irradiance[cte.YEAR][0]
    self.assertEqual(full_city_building_total_radiation, merged_city_building_total_radiation)

    for building in even_city.buildings:
      for surface in building.surfaces:
        surface.global_irradiance[cte.YEAR] = [3]

    merged_city = full_city.merge(even_city)
    first_merged_city_building_total_radiation = 0
    for building in merged_city.buildings:
      for surface in building.surfaces:
        if surface.global_irradiance:
          first_merged_city_building_total_radiation += surface.global_irradiance[cte.YEAR][0]
    merged_city = even_city.merge(full_city)
    second_merged_city_building_total_radiation = 0
    for building in merged_city.buildings:
      for surface in building.surfaces:
        if surface.global_irradiance:
          second_merged_city_building_total_radiation += surface.global_irradiance[cte.YEAR][0]
    self.assertAlmostEqual(first_merged_city_building_total_radiation, second_merged_city_building_total_radiation, 8)




"""
TestGeometryFactory test and validate the city model structure geometric parameters
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from unittest import TestCase

import hub.exports.exports_factory
from hub.helpers.dictionaries import MontrealFunctionToHubFunction
from hub.helpers.geometry_helper import GeometryHelper
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.geometry_factory import GeometryFactory


class TestGeometryFactory(TestCase):
  """
  Non-functional TestGeometryFactory
  Load testing
  """

  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()

  def _get_city(self, file, file_type, height_field=None, year_of_construction_field=None, function_field=None):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory(file_type,
                                 path=file_path,
                                 height_field=height_field,
                                 year_of_construction_field=year_of_construction_field,
                                 function_field=function_field,
                                 ).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  def _check_buildings(self, city):
    for building in city.buildings:
      self.assertIsNotNone(building.name, 'building name is none')
      self.assertIsNotNone(building.type, 'building type is none')
      self.assertIsNotNone(building.volume, 'building volume is none')
      self.assertIsNotNone(building.detailed_polyhedron, 'building detailed polyhedron is none')
      self.assertIsNotNone(building.simplified_polyhedron, 'building simplified polyhedron is none')
      self.assertIsNotNone(building.surfaces, 'building surfaces is none')
      self.assertIsNotNone(building.max_height, 'building max_height is none')
      self.assertEqual(len(building.external_temperature), 0, 'building external temperature is calculated')
      self.assertEqual(len(building.global_horizontal), 0, 'building global horizontal is calculated')
      self.assertEqual(len(building.diffuse), 0, 'building diffuse is calculated')
      self.assertEqual(len(building.beam), 0, 'building beam is calculated')
      self.assertIsNotNone(building.lower_corner, 'building lower corner is none')
      self.assertEqual(len(building.sensors), 0, 'building sensors are assigned')
      self.assertIsNotNone(building.internal_zones, 'no internal zones created')
      self.assertIsNotNone(building.grounds, 'building grounds is none')
      self.assertIsNotNone(building.walls, 'building walls is none')
      self.assertIsNotNone(building.roofs, 'building roofs is none')
      self.assertIsNotNone(building.internal_zones, 'building internal zones is none')
      for internal_zone in building.internal_zones:
        self.assertIsNone(internal_zone.usages, 'usage zones are defined')
        self.assertIsNone(internal_zone.thermal_archetype, 'thermal archetype is defined')
      self.assertIsNone(building.basement_heated, 'building basement_heated is not none')
      self.assertIsNone(building.attic_heated, 'building attic_heated is not none')
      self.assertIsNone(building.terrains, 'building terrains is not none')
      self.assertIsNone(building.average_storey_height, 'building average_storey_height is not none')
      self.assertIsNone(building.storeys_above_ground, 'building storeys_above_ground is not none')
      self.assertEqual(len(building.heating_demand), 0, 'building heating is not none')
      self.assertEqual(len(building.cooling_demand), 0, 'building cooling is not none')
      self.assertIsNotNone(building.eave_height, 'building eave height is none')
      self.assertIsNotNone(building.roof_type, 'building roof type is none')
      self.assertIsNotNone(building.floor_area, 'building floor_area is none')
      self.assertIsNone(building.households, 'building households is not none')
      self.assertFalse(building.is_conditioned, 'building is_conditioned is conditioned')

  def _check_surfaces(self, building):
    for surface in building.surfaces:
      self.assertIsNotNone(surface.name, 'surface name is none')
      self.assertIsNotNone(surface.id, 'surface id is none')
      self.assertIsNotNone(surface.lower_corner, 'surface envelope_lower_corner is none')
      self.assertIsNotNone(surface.upper_corner, 'surface envelope_upper_corner is none')
      self.assertIsNotNone(surface.perimeter_area, 'surface area_above_ground is none')
      self.assertIsNotNone(surface.azimuth, 'surface azimuth is none')
      self.assertIsNotNone(surface.inclination, 'surface inclination is none')
      self.assertIsNotNone(surface.type, 'surface type is none')
      self.assertEqual(len(surface.global_irradiance), 0, 'global irradiance is calculated')
      self.assertIsNotNone(surface.perimeter_polygon, 'surface perimeter_polygon is none')
      self.assertIsNone(surface.holes_polygons, 'surface hole_polygons is not none')
      self.assertIsNotNone(surface.solid_polygon, 'surface solid_polygon is none')
      self.assertIsNone(surface.short_wave_reflectance, 'surface short_wave_reflectance is not none')
      self.assertIsNone(surface.long_wave_emittance, 'surface long_wave_emittance is not none')
      self.assertIsNotNone(surface.inverse, 'surface inverse is none')
      self.assertEqual(len(surface.associated_thermal_boundaries), 0, 'associated_thermal_boundaries are assigned')
      self.assertIsNone(surface.vegetation, 'surface vegetation is not none')

  # citygml_classes
  def test_import_citygml(self):
    """
    Test city objects in the city
    :return: None
    """
    file = 'FZK_Haus_LoD_2.gml'
    city = self._get_city(file, 'citygml')
    self.assertTrue(len(city.buildings) == 1)
    self._check_buildings(city)
    for building in city.buildings:
      self._check_surfaces(building)
      city = ConstructionFactory('nrel', city).enrich()

  def test_import_obj(self):
    """
    Test obj import
    """
    file = 'kelowna.obj'
    city = self._get_city(file, 'obj')
    self.assertTrue(len(city.buildings) == 1)
    self._check_buildings(city)
    for building in city.buildings:
      self._check_surfaces(building)

  def test_import_geojson(self):
    """
    Test geojson import
    """
    file = Path(self._example_path / 'test.geojson').resolve()
    city = GeometryFactory('geojson',
                           path=file,
                           height_field='citygml_me',
                           year_of_construction_field='ANNEE_CONS',
                           aliases_field=['ID_UEV', 'CIVIQUE_DE', 'NOM_RUE'],
                           function_field='CODE_UTILI',
                           function_to_hub=MontrealFunctionToHubFunction().dictionary).city
    hub.exports.exports_factory.ExportsFactory('obj', city, self._output_path).export()
    for building in city.building_alias('01002777'):
      self.assertEqual('1', building.name, 'Wrong building name when looking for alias')
    self.assertEqual(8, len(city.building_alias('rue Sherbrooke Ouest  (MTL+MTO+WMT)')))
    self.assertEqual(17, len(city.buildings), 'wrong number of buildings')

    self.assertIsNotNone(city.city_object('15'), 'Building name 15 is missing in the city')
    city.remove_city_object(city.city_object('15'))
    self.assertIsNone(city.city_object('15'), 'Building name 15 wasn\'t removed')
    for building in city.buildings:
      _building = city.city_object(building.name)
      self.assertEqual(_building.name, building.name, 'hash map it\'s unsync')

  def test_map_neighbours(self):
    """
    Test neighbours map creation
    """
    file = 'test.geojson'

    city = self._get_city(file, 'geojson',
                          year_of_construction_field='ANNEE_CONS',
                          function_field='LIBELLE_UT')
    info_lod0 = GeometryHelper.city_mapping(city, plot=False)
    city = self._get_city(file, 'geojson',
                          height_field='citygml_me',
                          year_of_construction_field='ANNEE_CONS',
                          function_field='LIBELLE_UT')
    info_lod1 = GeometryHelper.city_mapping(city, plot=False)
    hub.exports.exports_factory.ExportsFactory('obj', city, self._output_path).export()
    self.assertEqual(info_lod0, info_lod1)
    self.assertEqual(2, len(city.city_object('1').neighbours))
    self.assertEqual(3, len(city.city_object('2').neighbours))
    self.assertEqual(2, len(city.city_object('3').neighbours))
    self.assertEqual(2, len(city.city_object('4').neighbours))
    self.assertEqual(3, len(city.city_object('5').neighbours))
    self.assertEqual(3, len(city.city_object('6').neighbours))
    self.assertEqual(1, len(city.city_object('8').neighbours))
    self.assertEqual(2, len(city.city_object('9').neighbours))
    self.assertEqual(2, len(city.city_object('10').neighbours))
    self.assertEqual(2, len(city.city_object('11').neighbours))
    self.assertEqual(2, len(city.city_object('12').neighbours))
    self.assertEqual(1, len(city.city_object('13').neighbours))
    self.assertEqual(2, len(city.city_object('14').neighbours))
    self.assertEqual(1, len(city.city_object('15').neighbours))
    self.assertEqual(1, len(city.city_object('16').neighbours))
    self.assertEqual(2, len(city.city_object('67').neighbours))
    self.assertEqual(1, len(city.city_object('68').neighbours))

    self.assertEqual('12', city.city_object('8').neighbours[0].name)
    self.assertEqual('14', city.city_object('13').neighbours[0].name)
    self.assertEqual('14', city.city_object('15').neighbours[0].name)



"""
Construction catalog window
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""


class Window:
  """
  Window class
  """
  def __init__(self, window_id, frame_ratio, g_value, overall_u_value, name, window_type=None):
    self._id = window_id
    self._frame_ratio = frame_ratio
    self._g_value = g_value
    self._overall_u_value = overall_u_value
    self._name = name
    self._type = window_type

  @property
  def id(self):
    """
    Get window id
    :return: str
    """
    return self._id

  @property
  def name(self):
    """
    Get window name
    :return: str
    """
    return self._name

  @property
  def frame_ratio(self):
    """
    Get window frame ratio
    :return: float
    """
    return self._frame_ratio

  @property
  def g_value(self):
    """
    Get thermal opening g-value
    :return: float
    """
    return self._g_value

  @property
  def overall_u_value(self):
    """
    Get thermal opening overall U-value in W/m2K
    :return: float
    """
    return self._overall_u_value

  @property
  def type(self):
    """
    Get transparent surface type, 'window' or 'skylight'
    :return: str
    """
    return self._type

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Window': {'id': self.id,
                          'name': self.name,
                          'type': self.type,
                          'frame ratio': self.frame_ratio,
                          'g-value': self.g_value,
                          'overall U value [W/m2K]': self.overall_u_value
                          }
               }
    return content



"""
Energy generation system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from __future__ import annotations
from typing import Union

from hub.city_model_structure.energy_systems.generic_generation_system import GenericGenerationSystem


class GenerationSystem:
  """
  GenerationSystem class
  """
  def __init__(self):
    self._heat_power = None
    self._cooling_power = None
    self._electricity_power = None
    self._storage_capacity = None
    self._generic_generation_system = None
    self._auxiliary_equipment = None

  @property
  def generic_generation_system(self) -> GenericGenerationSystem:
    """
    Get associated generic_generation_system
    :return: GenericGenerationSystem
    """
    return self._generic_generation_system

  @generic_generation_system.setter
  def generic_generation_system(self, value):
    """
    Set associated generic_generation_system
    :param value: GenericGenerationSystem
    """
    self._generic_generation_system = value

  @property
  def heat_power(self):
    """
    Get heat_power in W
    :return: float
    """
    return self._heat_power

  @heat_power.setter
  def heat_power(self, value):
    """
    Set heat_power in W
    :param value: float
    """
    self._heat_power = value

  @property
  def cooling_power(self):
    """
    Get cooling_power in W
    :return: float
    """
    return self._cooling_power

  @cooling_power.setter
  def cooling_power(self, value):
    """
    Set cooling_power in W
    :param value: float
    """
    self._cooling_power = value

  @property
  def electricity_power(self):
    """
    Get electricity_power in W
    :return: float
    """
    return self._electricity_power

  @electricity_power.setter
  def electricity_power(self, value):
    """
    Set electricity_power in W
    :param value: float
    """
    self._electricity_power = value

  @property
  def storage_capacity(self):
    """
    Get storage_capacity in J
    :return: float
    """
    return self._storage_capacity

  @storage_capacity.setter
  def storage_capacity(self, value):
    """
    Set storage_capacity in J
    :param value: float
    """
    self._storage_capacity = value

  @property
  def auxiliary_equipment(self) -> Union[None, GenerationSystem]:
    """
    Get auxiliary_equipment
    :return: GenerationSystem
    """
    return self._auxiliary_equipment

  @auxiliary_equipment.setter
  def auxiliary_equipment(self, value):
    """
    Set auxiliary_equipment
    :param value: GenerationSystem
    """
    self._auxiliary_equipment = value



"""
HvacTerminalUnit module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union


class HvacTerminalUnit:
  """
  HvacTerminalUnit class
  """
  def __init__(self):
    self._type = None

  @property
  def type(self) -> Union[None, str]:
    """
    Get type of hvac terminal unit defined for a thermal zone
    :return: None or str
    """
    return self._type

  @type.setter
  def type(self, value):
    """
    Set type of hvac terminal unit defined for a thermal zone
    :param value: str
    """
    if value is not None:
      self._type = str(value)



"""
Auth module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""

import bcrypt


class Auth:
  """
  Auth class
  """

  @staticmethod
  def hash_password(password: str) -> str:
    """
    Hashes a password
    :param password: the password to be hashed
    :return:
    """
    return bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt(14)).decode('utf-8')

  @staticmethod
  def check_password(password: str, hashed_password) -> bool:
    """
    Hashes a password
    :param password: the password to be checked
    :param hashed_password: the hashed password
    :return:
    """
    return bcrypt.checkpw(password.encode('utf-8'), hashed_password.encode('utf-8'))



"""
Construction catalog layer
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from hub.catalog_factories.data_models.construction.material import Material


class Layer:
  """
  Layer class
  """
  def __init__(self, layer_id, name, material, thickness):
    self._id = layer_id
    self._name = name
    self._material = material
    self._thickness = thickness

  @property
  def id(self):
    """
    Get layer id
    :return: str
    """
    return self._id

  @property
  def name(self):
    """
    Get layer name
    :return: str
    """
    return self._name

  @property
  def material(self) -> Material:
    """
    Get layer material
    :return: Material
    """
    return self._material

  @property
  def thickness(self):
    """
    Get layer thickness in meters
    :return: None or float
    """
    return self._thickness

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Layer': {'id': self.id,
                         'name': self.name,
                         'thickness [m]': self.thickness,
                         'material': self.material.to_dictionary()
                         }
               }
    return content
